#!/usr/bin/env python

"""
A really simple virtual processor employing a simple set of instructions which
ignore low-level operations and merely concentrate on variable access, structure
access, structure allocation and function invocations.

Copyright (C) 2007, 2008 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.

--------

The execution model of the virtual processor involves the following things:

  * Memory
  * PC (program counter) stack
  * Frame stack (containing invocation frames in use and in preparation plus
                temporary storage)
  * Local frame pointer stack
  * Invocation frame pointer stack
  * Exception handler stack
  * Registers: current value, boolean status value, source value, result,
               current exception

The memory contains constants, global variable references and program code.

The PC stack contains the return address associated with each function
invocation.

The frame pointer stack tracks the position of frames within the frame stack.
"""

class IllegalInstruction(Exception):
    pass

class IllegalAddress(Exception):
    pass

class EmptyPCStack(Exception):
    pass

class EmptyFrameStack(Exception):
    pass

class RSVPMachine:

    "A really simple virtual processor."

    def __init__(self, memory, objtable, paramtable, pc=None, debug=0):

        """
        Initialise the processor with a 'memory' (a list of values containing
        instructions and data) and the optional program counter 'pc'.
        """

        self.memory = memory
        self.objtable = objtable
        self.paramtable = paramtable
        self.pc = pc or 0
        self.debug = debug

        # Stacks.

        self.pc_stack = []
        self.frame_stack = []
        self.local_sp_stack = []
        self.invocation_sp_stack = []
        self.handler_stack = []

        # Registers.

        self.value = None
        self.status = None
        self.source = None
        self.result = None
        self.exception = None

    def load(self, address):

        "Return the value at the given 'address'."

        try:
            return self.memory[address]
        except IndexError:
            raise IllegalAddress, address

    def save(self, address, value):

        "Save to the given 'address' the specified 'value'."

        try:
            self.memory[address] = value
        except IndexError:
            raise IllegalAddress, address

    def new(self, size):

        """
        Allocate space of the given 'size', returning the address of the space.
        """

        addr = len(self.memory)
        for i in range(0, size):
            self.memory.append(None)
        return addr

    def push_pc(self, data):

        "Push 'data' onto the PC stack."

        self.pc_stack.append(data)

    def pull_pc(self):

        "Pull a value from the PC stack and return it."

        try:
            return self.pc_stack.pop()
        except IndexError:
            raise EmptyPCStack

    def run(self):

        "Execute code in the memory, starting from the current PC address."

        try:
            while 1:
                self.execute()
        except EmptyPCStack:
            pass

    def execute(self):

        "Execute code in the memory at the current PC address."

        instruction = self.load(self.pc).__class__.__name__
        if self.debug:
            print "%8d %s" % (self.pc, instruction)

        method = self.get_method(instruction)

        # Process any inputs of the instruction.

        self.process_inputs(instruction)
        method()

    def get_method(self, instruction):

        "Return the handler method for the given 'instruction'."

        method = getattr(self, instruction, None)
        if method is None:
            raise IllegalInstruction, (self.pc, instruction)
        return method

    def process_inputs(self, instruction):

        """
        Process any inputs of the 'instruction'. This permits any directly
        connected sub-instructions to produce the effects that separate
        instructions would otherwise have.
        """

        for input in (instruction.input, instruction.source):
            if input is not None:
                method = self.get_method(input)
                method()

    def jump(self, addr, next):

        """
        Jump to the subroutine at (or identified by) 'addr'. If 'addr'
        identifies a library function then invoke the library function and set
        PC to 'next' afterwards; otherwise, set PC to 'addr'.
        """

        if isinstance(addr, str):
            getattr(self, addr)()
            self.pc = next
        else:
            self.push_pc(self.pc + 2)
            self.pc = addr

    # Instructions.

    def LoadConst(self):
        op = self.load(self.pc)
        addr = op.get_operand()
        self.value = (None, addr)
        self.pc += 1

    def LoadName(self):
        n = self.load(self.pc).get_operand()
        frame = self.local_sp_stack[-1]
        self.value = self.frame_stack[frame + n]
        self.pc += 1

    def StoreName(self):
        n = self.load(self.pc).get_operand()
        frame = self.local_sp_stack[-1]
        self.frame_stack[frame + n] = self.value
        self.pc += 1

    LoadTemp = LoadName
    StoreTemp = StoreName

    def LoadAddress(self):
        addr = self.load(self.pc).get_operand()
        self.value = self.load(addr)
        self.pc += 1

    def LoadAddressContext(self):
        addr = self.load(self.pc).get_operand()
        value = self.load(addr)
        self.value = (self.value, value[1]) # replace the context with the current value
        self.pc += 1

    def StoreAddress(self):
        addr = self.load(self.pc).get_operand()
        self.save(addr, self.value)
        self.pc += 1

    def MakeObject(self):
        n = self.load(self.pc).get_operand()
        self.value = self.new(n)
        self.pc += 1

    def LoadAttr(self):
        n = self.load(self.pc).get_operand()
        context, ref = self.value
        self.value = self.load(ref + n)
        self.pc += 1

    def StoreAttr(self):
        n = self.load(self.pc).get_operand()
        context, ref = self.value
        self.save(ref + n, self.source)
        self.pc += 1

    def LoadAttrIndex(self): pass

    def StoreAttrIndex(self): pass

    def MakeFrame(self):
        n = self.load(self.pc).get_operand()
        self.invocation_sp_stack.append(len(self.frame_stack))
        self.frame_stack.extend([None] * n)
        self.pc += 1

    def JumpWithFrame(self):
        self.local_sp_stack.append(self.invocation_sp_stack[-1]) # adopt the invocation frame
        context, ref = self.value
        self.jump(ref.code_location, self.pc + 1) # return to the instruction after this one

    def DropFrame(self):
        result = self.pull()
        self.local_sp_stack.pop()
        frame = self.invocation_sp_stack.pop()
        self.frame_stack = self.frame_stack[:frame] # reset stack before call
        self.pc += 1

    def CheckFrame(self): pass

    def LoadCallable(self): pass

    def LoadContext(self):
        context, ref = self.value
        self.push((None, context))
        self.pc += 1

    def CheckSelf(self): pass

    def CheckContext(self): pass

    def Return(self):
        self.pc = self.pull_pc()

    def LoadResult(self):
        self.value = self.result

    def StoreResult(self):
        self.result = self.value

    def LoadException(self):
        self.value = self.exception

    def StoreException(self):
        self.exception = self.value

    def LoadBoolean(self):
        self.value = self.status

    def Jump(self):
        addr = self.load(self.pc).get_operand()
        self.pc = addr

    def JumpIfTrue(self):
        addr = self.load(self.pc).get_operand()
        if self.status:
            self.pc = addr
        else:
            self.pc += 1

    def JumpIfFalse(self):
        addr = self.load(self.pc).get_operand()
        if not self.status:
            self.pc = addr
        else:
            self.pc += 1

    def StoreFrame(self):
        pos = self.load(self.pc).get_operand()
        frame = self.invocation_sp_stack[-1] # different from the current frame after MakeFrame
        self.frame_stack[frame + pos] = self.value
        self.pc += 1

# vim: tabstop=4 expandtab shiftwidth=4
